import ssl
from abc import ABC, abstractmethod
from typing import Dict, Optional
from urllib.parse import urlencode

from aiohttp import ClientSession as Session
from multidict import istr

from .type import APPID, PT_QR_APP, Proxy

UA = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.74 Safari/537.36 Edg/99.0.1150.46"
CIPHERS = [
    "ECDHE+AESGCM",
    "ECDHE+CHACHA20",
    "DHE+AESGCM",
    "DHE+CHACHA20",
    "ECDH+AESGCM",
    "DH+AESGCM",
    "RSA+AESGCM",
    "!aNULL",
    "!eNULL",
    "!MD5",
    "!DSS",
]
XLOGIN_URL = "https://xui.ptlogin2.qq.com/cgi-bin/xlogin"


def ssl_context():
    c = ssl.create_default_context()
    c.set_ciphers(":".join(CIPHERS))
    return c


class LoginBase(ABC):
    login_sig: str = ""

    def __init__(self, sess: Session, app: APPID, proxy: Proxy, info: Optional[PT_QR_APP] = None):
        self.app = app
        self.proxy = proxy
        self.info = info if info else PT_QR_APP()
        sess.headers.update(
            {istr("DNT"): "1", istr("Referer"): "https://i.qq.com/", istr("User-Agent"): UA}
        )
        self.session = sess
        self.ssl = ssl_context()

    @property
    def xlogin_url(self):
        return (
            XLOGIN_URL
            + "?"
            + urlencode(
                {
                    "hide_title_bar": 1,
                    "style": 22,
                    "daid": self.app.daid,
                    "low_login": 0,
                    "qlogin_auto_login": 1,
                    "no_verifyimg": 1,
                    "link_target": "blank",
                    "appid": self.app.appid,
                    "target": "self",
                    "s_url": self.proxy.s_url,
                    "proxy_url": self.proxy.proxy_url,
                    "pt_qr_app": self.info.app,
                    "pt_qr_link": self.info.link,
                    "self_regurl": self.info.register,
                    "pt_qr_help_link": self.info.help,
                    "pt_no_auth": 1,
                }
            )
        )

    async def request(self):
        async with self.session.get(self.xlogin_url, ssl=self.ssl) as r:
            r.raise_for_status()
            self.local_token = int(r.cookies["pt_local_token"].value)
            self.login_sig = r.cookies["pt_login_sig"].value
        return self

    @abstractmethod
    async def login(self, *args, **kwds) -> Dict[str, str]:
        pass

    async def ja3Detect(self) -> dict:
        # for debuging
        async with self.session.get("https://ja3er.com/json", ssl=self.ssl) as r:
            return await r.json()

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        await self.session.close()



import asyncio
import base64
import json
import re
from math import floor
from random import choices, randint, random
from time import time
from typing import Any, Dict, Iterable, Tuple, cast
from urllib.parse import unquote, urlencode

from aiohttp import ClientSession as Session
from multidict import MutableMultiMapping, istr
from yarl import URL

from jssupport.execjs import ExecJS
from jssupport.jsjson import json_loads

from .jigsaw import Jigsaw

PREHANDLE_URL = "https://t.captcha.qq.com/cap_union_prehandle"
SHOW_NEW_URL = "https://t.captcha.qq.com/cap_union_new_show"
VERIFY_URL = "https://t.captcha.qq.com/cap_union_new_verify"

time_s = lambda: int(1e3 * time())
rnd6 = lambda: str(random())[2:8]


def hex_add(h: str, o: int):
    if h.endswith("#"):
        return h + str(o)
    if not h:
        return o
    return hex(int(h, 16) + o)[2:]


class ScriptHelper:
    def __init__(self, appid: int, sid: str, subsid: int) -> None:
        self.aid = appid
        self.sid = sid
        self.subsid = subsid

    def parseCaptchaConf(self, iframe: str):
        m = re.search(r"window\.captchaConfig=(\{.*\});", iframe)
        assert m
        ijson = m.group(1)
        self.conf: Dict[str, Any] = json_loads(ijson)  # type: ignore

    def cdn(self, cdn: int) -> str:
        assert cdn in (0, 1, 2)
        assert isinstance(self.conf, dict)
        data = {
            "aid": self.aid,
            "sess": self.conf["sess"],
            "sid": self.sid,
            "subsid": self.subsid + cdn,
        }
        if cdn:
            data["img_index"] = cdn
        path = self.conf[f"cdnPic{cdn if cdn else 1}"]
        return f"https://t.captcha.qq.com{path}?{urlencode(data)}"

    @staticmethod
    def slide_js_url(iframe):
        m = re.search(r"https://captcha\.gtimg\.com/1/tcaptcha-slide\.\w+\.js", iframe)
        assert m
        return m.group(0)

    @staticmethod
    def tdx_js_url(iframe):
        m = re.search(r'src="(/td[xc].js.*?)"', iframe)
        assert m
        return "https://t.captcha.qq.com" + m.group(1)


class VM:
    vmAvailable = 0
    vmByteCode = 0

    def __init__(self, tdx: str, header: MutableMultiMapping[str]) -> None:
        assert "User-Agent" in header
        assert "cookie" in header
        assert "Referer" in header
        self.tdx = ExecJS(js=self.constructWindow(header) + tdx)

    @staticmethod
    def constructWindow(header: MutableMultiMapping[str]):
        window = 'var href="%s",ua="%s",cookie="%s"\n' % (
            header["Referer"],
            header["User-Agent"],
            header["cookie"],
        )
        from pathlib import Path

        with open(Path(__file__).parent / "window.js") as f:
            window += f.read()
        return window

    async def getData(self):
        return unquote((await self.tdx("window.TDC.getData", True)).strip())

    async def getInfo(self) -> dict:
        return cast(dict, json_loads(await self.tdx("window.TDC.getInfo")))

    def setData(self, d: dict):
        self.tdx.addfunc("window.TDC.setData", d)

    def clearTc(self):
        return self.tdx("window.TDC.clearTc")

    async def getCookie(self):
        return (await self.tdx.get("window.document.cookie")).strip()


class Captcha:
    sess = None
    createIframeStart = 0
    subsid = 1

    # (c_login_2.js)showNewVC-->prehandle
    # prehandle(recall)--call tcapcha-frame.*.js-->new_show
    # new_show(html)--js in html->loadImg(url)
    def __init__(self, session: Session, ssl_context, appid: int, sid: str):
        self.session = session
        self.ssl = ssl_context
        self.appid = appid
        self.sid = sid

        self.session.headers.update({istr("referer"): "https://xui.ptlogin2.qq.com/"})

    @property
    def base64_ua(self):
        return base64.b64encode(self.session.headers["User-Agent"].encode()).decode()

    async def prehandle(self, xlogin_url):
        """
        call this before call `show`.

        args:
            xlogin_url: the url requested in QRLogin.request

        returns:
            dict

        example:
        ~~~
            {
                "state": 1,
                "ticket": "",
                "capclass": "1",
                "subcapclass": "15",
                "src_1": "cap_union_new_show",
                "src_2": "template/new_placeholder.html",
                "src_3": "template/new_slide_placeholder.html",
                "sess": "s0FXwryrkuYcdadBBb1d3_xihwN-42KXNAQQg6_5FKAuF-MdGmSlD9H7hQg29GLnk7uQLrsgHVRCQ7Mu1ylB6jY-XqrWaGPmoUfNJfJKjTY0ahaC16M6qb8bBsKJH67j8UPnI3r84TI35HMgtKh_t40jkHbp1l67l55rKEm4HHA27oIiEvD1rtUvb8UK2Rgfe7mb4wtuAvMrG-wVpZkamFhvx0e0GHlVCeDwBQ7o7cn0h4oH1V9pLN6GBkGiqBgHeTdFKqJH-FqNI*",
                "randstr": "",
                "sid": "874820883465668800"
            }
        ~~~
        """
        CALLBACK = "_aq_596882"
        data = {
            "aid": self.appid,
            "protocol": "https",
            "accver": 1,
            "showtype": "embed",
            "ua": self.base64_ua,
            "noheader": 1,
            "fb": 1,
            "enableDarkMode": 0,
            "sid": self.sid,
            "grayscale": 1,
            "clientype": 2,
            "cap_cd": "",
            "uid": "",
            "wxLang": "",
            "lang": "zh-CN",
            "entry_url": xlogin_url,
            # 'js': '/tcaptcha-frame.a75be429.js'
            "subsid": self.subsid,
            "callback": CALLBACK,
            "sess": "",
        }
        self.createIframeStart = time_s()
        async with self.session.get(PREHANDLE_URL, params=data, ssl=self.ssl) as r:
            r.raise_for_status()
            m = re.search(CALLBACK + r"\((\{.*\})\)", await r.text())

        assert m
        r = m.group(1)
        r = json.loads(r)
        self.sess = r["sess"]
        self.subsid = 2
        return r

    async def iframe(self):
        assert self.sess, "call prehandle first"

        data = {
            "aid": self.appid,
            "protocol": "https",
            "accver": 1,
            "showtype": "embed",
            "ua": self.base64_ua,
            "noheader": 1,
            "fb": 1,
            "enableDarkMode": 0,
            "sid": self.sid,
            "grayscale": 1,
            "clientype": 2,
            "sess": self.sess,
            "fwidth": 0,
            "wxLang": "",
            "tcScale": 1,
            "uid": "",
            "cap_cd": "",
            "subsid": self.subsid,
            "rnd": rnd6(),
            "prehandleLoadTime": time_s() - self.createIframeStart,
            "createIframeStart": self.createIframeStart,
        }
        async with self.session.get(SHOW_NEW_URL, params=data, ssl=self.ssl) as r:
            self.session.headers.update({istr("referer"): str(r.url)})
            self.prehandleLoadTime = data["prehandleLoadTime"]
            return await r.text()

    async def getBlob(self, iframe: str):
        js_url = ScriptHelper.slide_js_url(iframe)
        async with self.session.get(js_url, ssl=self.ssl) as r:
            r.raise_for_status()
            js = await r.text()
        m = re.search(r"'(!function.*;')", js)
        assert m
        return m.group(1)

    async def getTdx(self, iframe: str):
        js_url = ScriptHelper.tdx_js_url(iframe)
        header = self.session.headers.copy()
        header["cookie"] = "; ".join(
            f"{k}={v.value}"
            for k, v in self.session.cookie_jar.filter_cookies(URL("qq.com")).items()
        )
        async with self.session.get(js_url) as r:
            r.raise_for_status()
            self.vm = VM(await r.text(), header=header)

        m = re.search(r"TDC_itoken=([\w%]+);?", await self.vm.getCookie())
        assert m
        c = m.group(1)
        self.session.cookie_jar.update_cookies({"TDC_itoken": c})
        return self.vm

    async def matchMd5(self, iframe: str, powCfg: dict) -> Tuple[int, int]:
        if not hasattr(self, "_matchMd5"):
            blob = await self.getBlob(iframe)
            m = re.search(r",(function\(\w,\w,\w\).*?duration.*?),", blob)
            assert m
            blob = m.group(1)
            blob = f"var n=Object();!{blob}(null,n,null);"
            blob += "function matchMd5(p, m){return n.getWorkloadResult({nonce:p,target:m})}"
            self._matchMd5 = ExecJS(js=blob).bind("matchMd5")
        d = await self._matchMd5(powCfg["prefix"], powCfg["md5"])
        d = cast(Dict[str, int], json_loads(d.strip("\n")))
        return int(d["ans"]), int(d["duration"])

    @staticmethod
    def imitateDrag(x: int):
        assert x < 300
        # 244, 1247
        t = randint(1200, 1300)
        n = randint(50, 65)
        X = lambda i: randint(1, max(2, i // 10)) if i < n - 15 else randint(6, 12)
        Y = lambda: choices([-1, 1, 0], cum_weights=[0.1, 0.2, 1], k=1)[0]
        T = lambda: randint(*choices(((65, 280), (6, 10)), cum_weights=(0.05, 1), k=1)[0])
        xs = ts = 0
        drag = []
        for i in range(n):
            xi, ti = X(i), T()
            drag.append([xi, Y(), ti])
            xs += xi
            ts += ti
        drag.append([max(1, x - xs), Y(), max(1, t - ts)])
        drag.reverse()
        return drag

    async def rio(self, urls: Iterable[str]) -> Tuple[bytes, ...]:
        async def inner(url):
            async with self.session.get(url, ssl=self.ssl) as r:
                r.raise_for_status()
                return await r.content.read()

        return await asyncio.gather(*(inner(i) for i in urls))

    async def verify(self):
        """
        example:
        ~~~
        {
            "errorCode": "0",
            "randstr": "@VTn",
            "ticket": "t03UtRJOy9txaidDDdx5FBzSN_uwipfzMGe1pjDMIoO3dS2UUp1EEWiuZmIotD_709cAYhPGWo2M-uQZxorFH8JtGEhqpSYeRQ1h84opX2TQYWGRFATffLj8vsw_U3YJPHR5MPcvHsVGtM*",
            "errMessage": "",
            "sess": ""
        }
        ~~~
        """
        s = ScriptHelper(self.appid, self.sid, self.subsid)
        iframe = await self.iframe()
        s.parseCaptchaConf(iframe)
        Ians, duration = await self.matchMd5(iframe, s.conf["powCfg"])
        await self.getTdx(iframe)

        waitEnd = time() + 0.6 * random() + 0.9

        j = Jigsaw(*await self.rio(s.cdn(i) for i in range(3)), top=floor(int(s.conf["spt"])))

        self.vm.setData({"clientType": 2})
        self.vm.setData({"coordinate": [10, 24, 0.4103]})
        self.vm.setData(
            {
                "trycnt": 1,
                "refreshcnt": 0,
                "slideValue": self.imitateDrag(floor(j.left * j.rate)),
                "dragobj": 1,
            }
        )
        self.vm.setData({"ft": "qf_7P_n_H"})
        collect = await self.vm.getData()
        data = {
            "aid": self.appid,
            "protocol": "https",
            "accver": 1,
            "showtype": s.conf["showtype"],
            "ua": self.base64_ua,
            "noheader": s.conf["noheader"],
            "fb": 1,
            "enableDarkMode": 0,
            "sid": self.sid,
            "grayscale": 1,
            "clientype": 2,
            "sess": s.conf["sess"],
            "fwidth": 0,
            "wxLang": "",
            "tcScale": 1,
            "uid": s.conf["uid"],
            "cap_cd": "",
            "rnd": rnd6(),
            "prehandleLoadTime": self.prehandleLoadTime,
            "createIframeStart": self.createIframeStart,
            "subsid": self.subsid,
            "cdata": 0,
            "vsig": s.conf["vsig"],
            "websig": s.conf["websig"],
            "subcapclass": s.conf["subcapclass"],
            "fpinfo": "",
            "ans": f"{j.left},{j.top};",
            "nonce": s.conf["nonce"],
            "vlg": f"{self.vm.vmAvailable}_{self.vm.vmByteCode}_1",
            "pow_answer": hex_add(s.conf["powCfg"]["prefix"], Ians) if Ians else Ians,
            "pow_calc_time": duration,
            "eks": (await self.vm.getInfo())["info"],
            "tlg": len(collect),
            s.conf["collectdata"]: collect,
            # TODO: unknown
            # 'vData': 'gC*KM-*rjuHBcUjIt9kL6SV6JGdgfzMmP0BiFcaDg_7ctHwCjeoz4quIjb2FTgdJLBeCcKCZB_Mv7suXumolfmpSKZVIp7Un2N3b*fbwHX9aqRgjp5fmsgkf6aOgnhU_ttr_4xJZKVjStGX*hMwgBeHE_zuz-iDKy1coGdurLh559T6MoBdJdMAxtIlGJxAexbt6eDz3Aw5pD_tR01ElO7YY',
        }
        await asyncio.sleep(max(0, waitEnd - time()))
        async with self.session.post(VERIFY_URL, data=data, ssl=self.ssl) as r:
            self.sess = None
            self.createIframeStart = 0
            self.prehandleLoadTime = 0
            r = await r.json()

        if int(r["errorCode"]):
            raise RuntimeError(f"Code {r['errorCode']}: {r['errMessage']}")
        return r



from typing import Dict

from aiohttp import ClientResponse
from aiohttp.client_exceptions import ClientResponseError


def raise_for_status(response: ClientResponse, *accept_code: int):
    """A checker more strict than `ClientResponse.raise_for_status`.

    Args:
        response (ClientResponse): Client Response. Default is None, means `(200, )`

    Raises:
        `ClientResponseError`: if status not in `accept_code`
    """
    accept_code = accept_code or (200,)
    if response.status not in accept_code:
        raise ClientResponseError(
            response.request_info,
            response.history,
            status=response.status,
            message=response.reason or "",
            headers=response.headers,
        )


def get_all_cookie(response: ClientResponse) -> Dict[str, str]:
    cookies = response.headers.getall("Set-Cookie")
    cookie_kv = (i.split(";")[0].split("=", maxsplit=1) for i in cookies)
    d = {}
    for t in cookie_kv:
        if len(t) < 2:
            continue
        k, v = t
        d[k] = d.get(k) or v
    return d



"""
Basic wrapper of Qzone HTTP interface.
"""

import logging
from functools import wraps
from random import randint, random
from typing import Callable, Dict, List, Optional, Tuple, Union, cast
from urllib.parse import parse_qs, quote, unquote

from aiohttp import ClientSession as Session
from aiohttp.client_exceptions import ClientResponseError
from multidict import istr

import aioqzone.api.constant as const
from aioqzone.utils.daug import ud
from jssupport.jsjson import JsonValue, json_loads
from qqqr.base import UA
from qqqr.utils import raise_for_status

from ..exception import CorruptError, QzoneError
from ..interface.login import Loginable
from ..type.internal import AlbumData, LikeData
from ..utils.regex import response_callback
from ..utils.time import time_ms

logger = logging.getLogger(__name__)

StrDict = Dict[str, JsonValue]


class QzoneApi:
    """Just a wrapper for Qzone http interface. No type validating."""

    encoding = "utf-8"
    host = "https://user.qzone.qq.com"

    def __init__(self, session: Session, loginman: Loginable, ua: str = UA) -> None:
        session.headers.update({istr("User-Agent"): ua})
        self.sess = session
        self.login = loginman

    async def _get_gtk(self) -> int:
        """Get gtk with async-lock

        :raises `aioqzone.exception.QzoneError`: if gtk is 0
        :return: gtk
        """
        async with self.login.lock:
            gtk = self.login.gtk
        if gtk == 0:
            raise QzoneError(-3000)
        return gtk

    async def aget(self, url: str, params: Optional[Dict[str, str]] = None):
        params = params or {}
        params = ud(params, {"g_tk": str(await self._get_gtk())})
        self.sess.headers.update(
            {istr("referer"): f"https://user.qzone.qq.com/{self.login.uin}/infocenter"}
        )
        return self.sess.get(self.host + url, params=params)

    async def apost(
        self, url: str, params: Optional[Dict[str, str]] = None, data: Optional[dict] = None
    ):
        params = params or {}
        params = ud(params, {"g_tk": str(await self._get_gtk())})
        self.sess.headers.update(
            {istr("referer"): f"https://user.qzone.qq.com/{self.login.uin}/infocenter"}
        )
        return self.sess.post(self.host + url, params=params, data=data)

    def _relogin_retry(self, func: Callable):
        """A decorator which will relogin and retry given func if cookie expired.

        'cookie expired' is indicated by:
        1. `aioqzone.exception.QzoneError` code -3000 or -4002
        2. HTTP response code 403

        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        .. note:: Decorate code as less as possible
        .. warning:: Do NOT modify args in the wrapped code.
        """

        @wraps(func)
        async def relogin_wrapper(*args, **kwds):
            try:
                return await func(*args, **kwds)
            except QzoneError as e:
                if e.code not in [-3000, -4002]:
                    raise e
            except ClientResponseError as e:
                if e.status != 403:
                    raise e

            logger.info(f"Cookie expire in {func.__qualname__}. Relogin...")
            cookie = await self.login.new_cookie()
            return await func(*args, **kwds)

        return relogin_wrapper

    def _rtext_handler(
        self,
        rtext: str,
        cb: bool = True,
        errno_key: Tuple[str, ...] = ("code", "err"),
        msg_key: Tuple[str, ...] = ("msg", "message"),
    ) -> StrDict:
        """Deal with rtext from Qzone api response, returns parsed json dict.
        Inner used only.

        :param rtext: result from :external:meth:`asyncio.ClientResponse.text`
        :param cb: The text is to be parsed by callback_regex, defaults to True.
        :param errno_key: Error # key, defaults to ('code', 'err').
        :param msg_key: Error message key, defaults to ('msg', 'message').

        :raises `aioqzone.exception.QzoneError`: if errno != 0
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: json response
        """
        if cb:
            match = response_callback.search(rtext)
            assert match
            rtext = match.group(1)
        r = json_loads(rtext)
        assert isinstance(r, dict)

        err = next(filter(lambda i: i is not None, (r.get(i) for i in errno_key)), None)
        assert err is not None
        assert isinstance(err, (int, str))
        err = int(err)

        if err != 0:
            msg = next(filter(None, (r.get(i) for i in msg_key)), None)
            if msg:
                raise QzoneError(err, msg, rdict=r)
            else:
                raise QzoneError(err, rdict=r)
        return r  # type: ignore

    class FeedsMoreTransaction:
        """Represents a feeds3_html_more transaction."""

        def __init__(self, default: Optional[Dict[int, str]] = None) -> None:
            self.extern = default or {}

        def parse(self, page: int):
            unquoted = self.extern.get(page, "undefined")
            if unquoted == "undefined":
                return {}
            return {k: v[-1] for k, v in parse_qs(unquoted, keep_blank_values=True).items()}

    async def feeds3_html_more(
        self, pagenum: int, trans: Optional[FeedsMoreTransaction] = None, count: int = 10
    ) -> StrDict:
        """return a list of dict, each dict reps a page of feeds.

        :param pagenum: #page >= 0
        :param trans: reps a skim transaction. Mutable.
        :param count: feed count, defaults to 10.

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: feed attributes and html feed
        """

        default = {
            "scope": 0,
            "view": 1,
            "daylist": "",
            "uinlist": "",
            "gid": "",
            "flag": 1,
            "filter": "all",
            "applist": "all",
            "refresh": 0,
            "aisortEndTime": 0,
            "aisortOffset": 0,
            "getAisort": 0,
            "aisortBeginTime": 0,
            "firstGetGroup": 0,
            "icServerTime": 0,
            "mixnocache": 0,
            "scene": 0,
            "dayspac": "undefined",
            "sidomain": "qzonestyle.gtimg.cn",
            "useutf8": 1,
            "outputhtmlfeed": 1,
        }
        trans = trans or self.FeedsMoreTransaction()
        query = {
            "rd": random(),
            "uin": self.login.uin,
            "pagenum": pagenum + 1,
            "begintime": trans.parse(pagenum).get("basetime", "undefined"),
            "count": count,
            "usertime": time_ms(),
            "externparam": quote(trans.extern.get(pagenum, "undefined")),
        }

        @self._relogin_retry
        async def retry_closure():
            async with await self.aget(const.feeds3_html_more, ud(default, query)) as r:
                r.raise_for_status()
                rtext = await r.text(encoding=self.encoding)

            return self._rtext_handler(rtext)

        r = await retry_closure()
        data = r["data"]
        trans.extern[pagenum + 1] = unquote(data["main"]["externparam"])  # type: ignore
        return cast(StrDict, data)

    async def emotion_getcomments(self, uin: int, tid: str, feedstype: int) -> StrDict:
        """Get complete html of a given feed

        :param uin: uin
        :param tid: feed id
        :param feedstype: feedstype in html

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: response dict
        """
        default = {
            "pos": 0,
            "num": 1,
            "cmtnum": 1,
            "t1_source": 1,
            "who": 1,
            "inCharset": "utf-8",
            "outCharset": "utf-8",
            "plat": "qzone",
            "source": "ic",
            "paramstr": 1,
            "fullContent": 1,
        }
        body = {
            "uin": uin,
            "tid": tid,
            "feedsType": feedstype,
            "qzreferrer": f"https://user.qzone.qq.com/{self.login.uin}",
        }

        @self._relogin_retry
        async def retry_closure():
            async with await self.apost(const.emotion_getcomments, data=ud(default, body)) as r:
                r.raise_for_status()
                rtext = await r.text(encoding=self.encoding)
            return self._rtext_handler(rtext)

        return await retry_closure()

    async def emotion_msgdetail(self, owner: int, fid: str) -> StrDict:
        """Get detail of a given msg.

        :param owner: owner uin
        :param fid: feed id, named fid, tid or feedkey

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: a dict reps the feed in detail

        .. note:: share msg is not support, i.e. `appid=311`
        """
        default = {
            "t1_source": 1,
            "ftype": 0,
            "sort": 0,
            "pos": 0,
            "num": 20,
            "callback": "callback",
            "code_version": 1,
            "format": "jsonp",
            "need_private_comment": 1,
        }
        query = {"uin": owner, "tid": fid}

        @self._relogin_retry
        async def retry_closure():
            async with await self.aget(const.emotion_msgdetail, params=ud(default, query)) as r:
                r.raise_for_status()
                rtext = await r.text(encoding=self.encoding)

            return self._rtext_handler(rtext)

        return await retry_closure()

    async def get_feeds_count(self) -> Dict[str, Union[int, list]]:
        """Get feeds update count (new feeds, new photos, new comments, etc)

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: update counts

        .. note::
            This api is also the 'keep-alive' signal to avoid cookie from expiring.
            Call this api every 300s can help keep cookie alive.
        """
        query = {"uin": self.login.uin, "rd": random()}

        @self._relogin_retry
        async def retry_closure():
            async with await self.aget(const.get_feeds_count, query) as r:
                r.raise_for_status()
                rtext = await r.text(encoding=self.encoding)

            return self._rtext_handler(rtext)

        r = await retry_closure()
        return r["data"]  # type: ignore

    async def like_app(self, likedata: LikeData, like: bool = True) -> bool:
        """Like or unlike a feed.

        :param likedata: Necessary data for like/unlike
        :param like: True as like, False as unlike, defaults to True.

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: success flag

        .. versionchanged:: 0.2.7
            not a @noexcept method.
        """
        default = {
            "from": 1,
            "active": 0,
            "fupdate": 1,
        }
        body = {
            "qzreferrer": f"https://user.qzone.qq.com/{self.login.uin}",
            "opuin": self.login.uin,
            "unikey": likedata.unikey,
            "curkey": likedata.curkey,
            "appid": likedata.appid,
            "typeid": likedata.typeid,
            "abstime": likedata.abstime,
            "fid": likedata.fid,
        }
        url = const.internal_dolike_app if like else const.internal_unlike_app

        @self._relogin_retry
        async def retry_closure():
            async with await self.apost(url, data=ud(default, body)) as r:
                r.raise_for_status()
                rtext = await r.text(encoding=self.encoding)
            return self._rtext_handler(rtext, errno_key=("code", "ret"))

        try:
            await retry_closure()
            return True
        except QzoneError as e:
            logger.warning(f"Error in dolike/unlike. {e}")
            return False
        except ClientResponseError as e:
            logger.error("Error in dolike/unlike.", exc_info=True)
            raise e
        except SystemExit as e:
            raise e  # pass through SystemExit
        except BaseException as e:
            logger.fatal("Uncaught Exception in dolike/unlike!", exc_info=True)
            exit(1)

    async def floatview_photo_list(self, album: AlbumData, num: int) -> StrDict:
        """Get detail of an album, including raw image url.

        :param album: Necessary album data
        :param num: pic num

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error
        :raises `aioqzone.exception.CorruptError`: maybe data is corruptted

        :return: album details

        .. warning::
            This api is a 'slow' api, means it always returns an error code,
            requesting a retry a moment later. Use :meth:`.emotion_msgdetail` to get raw photo/video in
            a feed.
        """

        default = {
            "callback": "viewer_Callback",
            "cmtOrder": 1,
            "fupdate": 1,
            "plat": "qzone",
            "source": "qzone",
            "cmtNum": 0,
            "likeNum": 0,
            "inCharset": "utf-8",
            "outCharset": "utf-8",
            "callbackFun": "viewer",
            "offset": 0,
            "appid": 311,
            "isFirst": 1,
            "need_private_comment": 1,
            "shootTime": "",
        }
        query = {
            "topicId": album.topicid,
            "picKey": album.pickey,
            "hostUin": album.hostuin,
            "number": num,
            "uin": self.login.uin,
            "_": time_ms(),
            "t": randint(int(1e8), int(1e9 - 1))
            # The distribution is not consistent with photo.js; but the format is.
        }

        @self._relogin_retry
        async def retry_closure():
            async with await self.aget(const.floatview_photo_list, ud(default, query)) as r:
                r.raise_for_status()
                rtext = await r.text()
            return self._rtext_handler(rtext)

        rjson = (await retry_closure())["data"]
        if query["t"] != int(rjson.pop("t")):  # type: ignore
            raise CorruptError("Something unexpected occured in transport.")
        return rjson  # type: ignore

    async def emotion_msglist(
        self,
        uin: int,
        num: int = 20,
        pos: int = 0,
    ) -> List[StrDict]:
        """Get msg(feed?) list of a user.

        :param uin: uin
        :param num: number, defaults to 20
        :param pos: start position, defaults to 0

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: a list of messages

        .. versionadded:: 0.2.6
        """
        add = {
            "hostUin": uin,
            "notice": 0,
            "inCharset": "utf-8",
            "outCharset": "utf-8",
            "cgi_host": "https://user.qzone.qq.com" + const.emotion_msglist,
        }
        param = {
            "uin": uin,
            "ftype": 0,
            "sort": 0,
            "pos": pos,
            "num": num,
            "replynum": 0,
            "callback": "callback",
            "code_version": 1,
            "format": "jsonp",
            "need_private_comment": 1,
        }

        @self._relogin_retry
        async def retry_closure():
            async with await self.aget(
                const.emotion_msglist, ud(param, add) if pos else param
            ) as r:
                r.raise_for_status()
                rtext = await r.text()
            return self._rtext_handler(rtext)

        data = await retry_closure()
        return data["msglist"]  # type: ignore

    async def emotion_publish(self, content: str, right: int = 0) -> StrDict:
        """Publish a feed. appid=311.

        :param content: text content.
        :param right: feed access right, defaults to 0 (Not used till now)

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: qzone response as is, containing feed html and fid.

        .. versionadded:: 0.2.6
        .. warning:: This api is under development. It has basic functions only.
        """
        default = {
            "syn_tweet_verson": 1,
            "paramstr": 1,
            "pic_template": "",
            "richtype": "",
            "richval": "",
            "special_url": "",
            "subrichtype": "",
            "who": 1,
            "ver": 1,
            "to_sign": 0,
            "code_version": 1,
            "format": "fs",
        }
        body = {
            "ugc_right": 64,
            "con": content,
            "feedversion": 1,
            "hostuin": self.login.uin,
            "qzreferrer": f"https://user.qzone.qq.com/{self.login.uin}",
        }

        @self._relogin_retry
        async def retry_closure():
            async with await self.apost(const.emotion_publish, data=ud(default, body)) as r:
                r.raise_for_status()
                rtext = await r.text()
            return self._rtext_handler(rtext)

        return await retry_closure()

    async def emotion_delete(
        self,
        fid: str,
        abstime: int,
        appid: int,
        typeid: int,
        topicId: str,
        uin: Optional[int] = None,
    ) -> Optional[StrDict]:
        """Delete a feed.

        :param fid: feed id, named tid, feedkey, etc.
        :param abstime: feed create time. `now` in :meth:`.emotion_publish` result dict.
        :param appid: appid
        :param typeid: typeid
        :param topicId: topic id, got from html
        :param uin: host uin, defaults to None, means current logined user.

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: qzone response as is, usually nothing meaningful.

        .. versionadded:: 0.2.6
        """
        body = {
            "uin": uin or self.login.uin,
            "topicId": topicId,
            "feedsType": typeid,
            "feedsFlag": 0,
            "feedsKey": fid,
            "feedsAppid": appid,
            "feedsTime": abstime,
            "fupdate": 1,
            "ref": "feeds",
            "qzreferrer": f"https://user.qzone.qq.com/{self.login.uin}",
        }

        @self._relogin_retry
        async def retry_closure():
            async with await self.apost(const.emotion_delete, data=body) as r:
                # upstream error: qzone server returns 503, but the feed operation is done.
                raise_for_status(r, 200, 503)
                if r.status == 503:
                    return
                rtext = await r.text()
            return self._rtext_handler(rtext)

        return await retry_closure()

    async def emotion_update(self, fid: str, content: str, uin: Optional[int] = None) -> StrDict:
        """Update content of a feed.

        :param fid: feed id, named feedkey, tid, etc.
        :param content: new content in text.
        :param uin: host uin, defaults to None, means current logined user.

        :raises `aiohttp.ClientResponseError`: error http response code
        :raises `aioqzone.exception.QzoneError`: error qzone response code
        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: qzone response as is.

        .. versionadded:: 0.2.6
        .. warning:: This api is under development. It has basic functions only.
        """
        default = {
            "syn_tweet_verson": 1,
            "paramstr": 1,
            "pic_template": "",
            "richtype": "",
            "richval": "",  # album attribute comma list
            "special_url": "",
            "subrichtype": "",
            "feedversion": 1,
            "ver": 1,
            "code_version": "1",
            "format": "fs",
        }
        body = {
            "tid": fid,
            "con": content,
            "ugc_right": 64,
            "to_sign": 0,
            "ugcright_id": "TODO",  # TODO
            "hostuin": uin or self.login.uin,
            "qzreferrer": f"https://user.qzone.qq.com/{self.login.uin}",
            # 'pic_bo': ''
        }

        @self._relogin_retry
        async def retry_closure():
            async with await self.apost(const.emotion_update, data=ud(default, body)) as r:
                r.raise_for_status()
                rtext = await r.text()
            return self._rtext_handler(rtext)

        return await retry_closure()



from dataclasses import dataclass


@dataclass(frozen=True)
class PT_QR_APP:
    app: str = ""
    link: str = ""
    register: str = ""
    help: str = ""


@dataclass(frozen=True)
class Proxy:
    proxy_url: str
    s_url: str


@dataclass(frozen=True)
class APPID:
    appid: int
    daid: int


@dataclass
class CheckResult:
    code: int
    verifycode: str
    salt: str
    verifysession: str
    isRandSalt: int
    ptdrvs: str
    session: str

    def __post_init__(self):
        self.code = int(self.code)
        self.isRandSalt = int(self.isRandSalt)
        salt = self.salt.split(r"\x")[1:]
        salt = [chr(int(i, 16)) for i in salt]
        self.salt = "".join(salt)



"""
Offers a class to execute js by communicating with subprocess.

... warning:: On win32 platform, importing this module will change asyncio event loop policy to
:external:class:`asyncio.WindowsProactorEventLoopPolicy`!
"""

import asyncio
from collections import defaultdict
from functools import partial
from sys import platform
from typing import Any, Coroutine, Optional


class ExecJS:
    def __init__(self, node: str = "node", *, js: Optional[str] = None):
        self.js = js
        self.que = []
        self.node = node
        assert self.version() is not None, f"`{self.node}` is not installed."
        assert self.loop_policy_check()

    @staticmethod
    def callstr(func, *args, asis: bool = False) -> str:
        tostr = {bool: lambda i: ["false", "true"][i]}
        tostr = defaultdict(lambda: repr, tostr)
        quoted = (repr(i) if asis else tostr[type(i)](i) for i in args)
        return f'{func}({",".join(quoted)})'

    def addfunc(self, func: str, *args):
        self.que.append((func, *args))

    async def _exec(self, js: str):
        p = await asyncio.subprocess.create_subprocess_exec(
            self.node,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await p.communicate(js.encode())
        removesuffix = lambda s: s[:-1] if str.endswith(s, "\n") else s
        if stderr:
            raise RuntimeError(removesuffix(stderr.decode()))
        return removesuffix(stdout.decode())

    def __call__(self, func: str, *args, asis: bool = False) -> Coroutine[Any, Any, str]:
        js = self.js
        assert js is not None
        for i in self.que:
            js += self.callstr(*i, asis=asis)
            js += ";"
        self.que.clear()
        js += f"\nconsole.log({self.callstr(func, *args, asis=asis)});process.exit();"
        return self._exec(js)

    def get(self, prop: str):
        assert self.js is not None
        js = self.js + f"\nconsole.log({prop});process.exit();"
        return self._exec(js)

    def bind(self, func: str, new: bool = True):
        n = ExecJS(self.node, js=self.js) if new else self
        return partial(n.__call__, func)

    def version(self):
        import subprocess

        try:
            p = subprocess.Popen(
                [self.node, "-v"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except FileNotFoundError:
            return
        stdout, stderr = p.communicate()
        if stderr:
            raise RuntimeError(stderr.decode())
        return stdout.decode()

    def loop_policy_check(self):
        """On Windows, the default event loop :external:class:`asyncio.ProactorEventLoop` supports subprocesses,
        whereas :external:class:`asyncio.SelectorEventLoop` does not.
        """
        if platform == "win32" and isinstance(
            asyncio.get_event_loop_policy(), asyncio.WindowsSelectorEventLoopPolicy
        ):
            return False
        return True


if platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())



def hash33(key: str, phash: int = 0):
    # str.hash33
    for c in key:
        phash += (phash << 5) + ord(c)
    return 0x7FFFFFFF & phash


def gtk(p_skey: str):
    return hash33(p_skey, phash=5381)



"""
Use this module to get some data from Qzone html feed
"""
import logging
import re
from typing import Iterable, List, Optional, Union, cast

from lxml.html import HtmlElement, fromstring
from pydantic import BaseModel, HttpUrl

from ..type.entity import RespEntities, TextEntity
from ..type.internal import AlbumData
from ..type.resp import PicRep
from ..utils.daug import di

logger = logging.getLogger(__name__)


class HtmlInfo(BaseModel):
    """Some info that must be extract from html response"""

    feedstype: int
    topicid: str
    complete: bool
    unikey: Optional[Union[HttpUrl, str]] = None
    curkey: Optional[Union[HttpUrl, str]] = None
    islike: int = 0

    @classmethod
    def from_html(cls, html: str):
        root: HtmlElement = fromstring(html)
        safe = lambda i: i[0] if i else HtmlElement()

        # feed data
        elm_fd = safe(root.cssselect('div.f-single-content i[name="feed_data"]'))

        # like data
        ffoot = safe(root.cssselect("div.f-single-foot"))
        likebtn = safe(ffoot.cssselect("a.qz_like_prase") + ffoot.cssselect("a.qz_like_btn_v3"))

        # cut toggle
        toggle = safe(root.cssselect('div.f-info a[data-cmd="qz_toggle"]'))

        return root, cls(
            feedstype=elm_fd.get("data-feedstype", None),
            topicid=elm_fd.get("data-topicid", None),
            complete=not len(toggle),
            unikey=likebtn.get("data-unikey", None),
            curkey=likebtn.get("data-curkey", None),
            islike=likebtn.get("data-islike", 0),
        )


class HtmlContent(BaseModel):
    entities: Optional[List[RespEntities]]
    pic: Optional[List[PicRep]] = None
    album: Optional[AlbumData] = None

    @classmethod
    def from_html(cls, html: Union[HtmlElement, str], hostuin: int = 0):
        root: HtmlElement = fromstring(html) if isinstance(html, str) else html
        mxsafe = lambda i: max(i, key=len) if i else HtmlElement()
        img_data = lambda a: {k[5:]: v for k, v in a.attrib.items() if k.startswith("data-")}

        def load_src(a: Iterable[HtmlElement]) -> Optional[HttpUrl]:
            o: Optional[HtmlElement] = next(filter(lambda i: i.tag == "img", a), None)
            if o is None:
                return
            src: str = o.get("src", "")
            if src.startswith("http"):
                return cast(HttpUrl, src)

            m = re.search(r"trueSrc:'(http.*?)'", o.get("onload", ""))
            if m:
                return cast(HttpUrl, m.group(1).replace("\\", ""))

            if "onload" in o.attrib:
                logger.warning("cannot parse @onload: " + o.get("onload", ""))
            elif "src" in o.attrib:
                logger.warning("cannot parse @src: " + o.get("src", ""))
            else:
                logger.warning(f"WTF is this? {dict(o.attrib)}")

        finfo = mxsafe(root.cssselect("div.f-info"))
        lia: List[HtmlElement] = root.cssselect("div.f-ct a.img-item")

        try:
            album = AlbumData.parse_obj(di(img_data(lia[0]), hostuin=hostuin))
        except:
            album = None

        pic = []
        for a in lia:
            src = load_src(a)  # type: ignore
            if not src:
                continue
            data = img_data(a)
            pic.append(
                PicRep(
                    height=data.get("height", 0),
                    width=data.get("width", 0),
                    url1=src,
                    url2=src,
                    url3=src,
                )
            )

        # TODO
        alltxt = TextEntity(type=2, con=finfo.text_content())
        return cls(entities=[alltxt], pic=pic, album=album)



import ast
import json
import logging
from platform import python_version_tuple
from textwrap import dedent
from typing import Callable, Dict, List, Union

logger = logging.getLogger(__name__)
JsonDict = Dict[Union[str, int], "JsonValue"]
JsonList = List["JsonValue"]
JsonValue = Union[bool, int, str, JsonDict, JsonList]


class NodeLoader:
    from .execjs import ExecJS

    jsonStringify = ExecJS(js="").bind("JSON.stringify", new=False)

    @classmethod
    async def json_loads(
        cls, js: str, try_load_first: bool = True, parser: Callable[[str], JsonValue] = json.loads
    ) -> JsonValue:
        """
        The json_loads function converts a string representation of JS/JSON data into a Python object.
        It may use :node:meth:`JSON.stringify` to convert js to json.

        :param js: Used to Pass in the json string.
        :param try_load_first: Used to Specify whether to try loading the json string with the `parser` first.
        :param parser: Used to Specify the function that will be used to parse the string.
        :return: A python object that represents the same content as the js/json string.
        """
        if try_load_first:
            try:
                return parser(js)
            except json.JSONDecodeError:
                pass

        json_str = await cls.jsonStringify(js, asis=True)
        try:
            return parser(json_str)
        except json.JSONDecodeError as e:
            logger.exception("Failed to decode json input!")
            logger.debug("json_str=%s", json_str)
            raise e


class AstLoader:
    class RewriteUndef(ast.NodeTransformer):
        def __init__(self) -> None:
            if int(python_version_tuple()[1]) < 8:
                # NOTE: visit_Constant not available on py37
                self.visit_Str = lambda node: ast.Str(s=node.s.replace("\\/", "/"))

        const = {
            "undefined": ast.Constant(value=None),
            "null": ast.Constant(value=None),
            "true": ast.Constant(value=True),
            "false": ast.Constant(value=False),
        }

        def visit_Name(self, node: ast.Name):
            if node.id in self.const:
                return self.const[node.id]
            return ast.Str(s=node.id)

        def visit_Constant(self, node: ast.Constant) -> Union[ast.Constant, ast.Str]:
            if not isinstance(node.value, str):
                return node
            return ast.Str(s=node.value.replace("\\/", "/"))

    @classmethod
    def json_loads(cls, js: str, filename: str = "stdin") -> JsonValue:
        """
        The json_loads function loads a JSON object from a js/json string. It uses standard
        :mod:`ast` module to parse the js/json.

        :param js: Used to Pass the js/json string to be parsed.
        :param filename: Used to Specify the name of the file that is being read. This is only for debug use.
        :rtype: `dict[str | int, Any]`
        :return: A jsonvalue object.
        """

        node = ast.parse(dedent(js), mode="eval")
        node = ast.fix_missing_locations(cls.RewriteUndef().visit(node))
        code = compile(node, filename, mode="eval")
        return eval(code)


def json_loads(js: str) -> JsonValue:
    """The json_loads function converts a string representation of JS/JSON data into a Python object.
    Current implementation is using :mod:`ast`.

    If you need more parameters or another implementation, call `xxxLoader.json_loads` instead.

    .. seealso:: :meth:`.AstLoader.json_loads`

    :param js: Used to Pass the JS/JSON string.
    :rtype: `dict[str | int, Any]`
    :return: A jsonvalue object.
    """
    return AstLoader.json_loads(js)



feeds3_html_more = "/proxy/domain/ic2.qzone.qq.com/cgi-bin/feeds/feeds3_html_more"
emotion_getcomments = "/proxy/domain/taotao.qzone.qq.com/cgi-bin/emotion_cgi_ic_getcomments"
emotion_msgdetail = "/proxy/domain/taotao.qq.com/cgi-bin/emotion_cgi_msgdetail_v6"
get_feeds_count = "/proxy/domain/ic2.qzone.qq.com/cgi-bin/feeds/cgi_get_feeds_count.cgi"
internal_dolike_app = "/proxy/domain/w.qzone.qq.com/cgi-bin/likes/internal_dolike_app"
internal_unlike_app = "/proxy/domain/w.qzone.qq.com/cgi-bin/likes/internal_unlike_app"
floatview_photo_list = "/proxy/domain/plist.photo.qq.com/fcgi-bin/cgi_floatview_photo_list_v2"
emotion_msglist = "/proxy/domain/taotao.qq.com/cgi-bin/emotion_cgi_msglist_v6"
emotion_publish = "/proxy/domain/taotao.qzone.qq.com/cgi-bin/emotion_cgi_publish_v6"
emotion_delete = "/proxy/domain/taotao.qzone.qq.com/cgi-bin/emotion_cgi_delete_v6"
emotion_update = (
    "https://user.qzone.qq.com/proxy/domain/taotao.qzone.qq.com/cgi-bin/emotion_cgi_update"
)



"""
date: 2021-07-20
"""

from enum import IntEnum

from .type import APPID, Proxy

QzoneProxy = Proxy(
    "https://qzs.qq.com/qzone/v6/portal/proxy.html",
    "https://qzs.qzone.qq.com/qzone/v5/loginsucc.html?para=izone",
)

QzoneAppid = APPID(549000912, 5)


class StatusCode(IntEnum):
    # Unified
    Authenticated = 0
    # QR
    Expired = 65
    Waiting = 66
    Scanned = 67
    # UP
    NeedCaptcha = 1
    WrongPassword = 3
    ForceQR = 10005
    NeedVerify = 10009  # Means, notify the owner
    RiskyNetwork = 23003



class TencentLoginError(RuntimeError):
    def __init__(self, code: int, msg: str, *args: object) -> None:
        self.code = code
        self.msg = msg
        super().__init__(*args)

    def __str__(self) -> str:
        return f"Code {self.code}: {self.msg}"


class UserBreak(KeyboardInterrupt):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



"""
Collect some built-in login manager w/o caching.
Users can inherit these managers and implement their own caching logic.
"""

import logging
from enum import Enum
from typing import Dict, List, Optional, Type

from aiohttp import ClientSession

from qqqr.constants import QzoneAppid, QzoneProxy, StatusCode
from qqqr.exception import TencentLoginError, UserBreak
from qqqr.qr import QRLogin
from qqqr.up import UPLogin, User

from ..exception import LoginError
from ..interface.hook import QREvent
from ..interface.login import Loginable

logger = logging.getLogger(__name__)


class ConstLoginMan(Loginable):
    """Only for test"""

    def __init__(self, uin: int, cookie: dict) -> None:
        super().__init__(uin)
        self._cookie = cookie

    async def _new_cookie(self) -> Dict[str, str]:
        return self._cookie


class UPLoginMan(Loginable):
    def __init__(self, sess: ClientSession, uin: int, pwd: str) -> None:
        Loginable.__init__(self, uin)
        self.sess = sess
        self._pwd = pwd

    async def _new_cookie(self) -> Dict[str, str]:
        """
        :raises `qqqr.exception.TencentLoginError`: login error when up login.
        :raises `SystemExit`: if unexpected error raised
        """
        try:
            login = UPLogin(self.sess, QzoneAppid, QzoneProxy, User(self.uin, self._pwd))
            cookie = await login.login(await login.check())
            self.add_hook_ref("hook", self.hook.LoginSuccess())
            self.sess.cookie_jar.update_cookies(cookie)
            return cookie
        except TencentLoginError as e:
            self.add_hook_ref("hook", self.hook.LoginFailed(e.msg or ""))
            logger.warning(str(e))
            raise e
        except NotImplementedError as e:
            self.add_hook_ref("hook", self.hook.LoginFailed("10009：需要手机验证"))
            logger.warning(str(e))
            raise TencentLoginError(StatusCode.NeedVerify, "Dynamic code verify not implemented")
        except:
            logger.fatal("Unexpected error in QR login.", exc_info=True)
            try:
                await self.hook.LoginFailed("密码登录期间出现奇怪的错误😰请检查日志以便寻求帮助.")
            finally:
                exit(1)


class QRLoginMan(Loginable):
    hook: QREvent

    def __init__(self, sess: ClientSession, uin: int, refresh_time: int = 6) -> None:
        Loginable.__init__(self, uin)
        self.sess = sess
        self.refresh = refresh_time

    async def _new_cookie(self) -> Dict[str, str]:
        """
        :raises `qqqr.exception.UserBreak`: qr polling task is canceled
        :raises `TimeoutError`: qr polling task timeout
        :raises `SystemExit`: if unexpected error raised when polling
        """
        assert isinstance(self.hook, QREvent)

        man = QRLogin(self.sess, QzoneAppid, QzoneProxy)
        task = man.loop(send_callback=self.hook.QrFetched, refresh_time=self.refresh)

        async def tmp_cancel():
            task.cancel()

        async def tmp_resend():
            await self.hook.QrFetched(await man.show())  # must be sent at once

        self.hook.cancel = tmp_cancel
        self.hook.resend = tmp_resend

        try:
            cookie = await task
            self.add_hook_ref("hook", self.hook.QrSucceess())
            self.add_hook_ref("hook", self.hook.LoginSuccess())
            self.sess.cookie_jar.update_cookies(cookie)
            return cookie
        except TimeoutError as e:
            await self.hook.QrFailed()
            logger.warning(str(e))
            self.add_hook_ref("hook", self.hook.QrFailed(str(e)))
            self.add_hook_ref("hook", self.hook.LoginFailed(str(e)))
            raise e
        except KeyboardInterrupt as e:
            raise UserBreak from e
        except:
            logger.fatal("Unexpected error in QR login.", exc_info=True)
            msg = "二维码登录期间出现奇怪的错误😰请检查日志以便寻求帮助."
            try:
                await self.hook.QrFailed(msg)
                await self.hook.LoginFailed(msg)
            finally:
                exit(1)
        finally:
            self.hook.cancel = self.hook.resend = None


class MixedLoginMan(UPLoginMan, QRLoginMan):
    class QrStrategy(str, Enum):
        force = "force"
        prefer = "prefer"
        allow = "allow"
        forbid = "forbid"

    def __init__(
        self,
        sess: ClientSession,
        uin: int,
        strategy: QrStrategy,
        pwd: Optional[str] = None,
        refresh_time: int = 6,
    ) -> None:
        self.strategy = strategy
        if strategy != "force":
            assert pwd
            UPLoginMan.__init__(self, sess, uin, pwd)
        if strategy != "forbid":
            QRLoginMan.__init__(self, sess, uin, refresh_time)

    async def _new_cookie(self) -> Dict[str, str]:
        """

        :raises `qqqr.exception.UserBreak`: qr login canceled
        :raises `aioqzone.exception.LoginError`: not logined
        :raises `SystemExit`: unexcpected error

        :return: cookie
        """
        order: List[Type[Loginable]] = {
            "force": [QRLoginMan],
            "prefer": [QRLoginMan, UPLoginMan],
            "allow": [UPLoginMan, QRLoginMan],
            "forbid": [UPLoginMan],
        }[self.strategy]
        for c in order:
            try:
                return await c._new_cookie(self)
            except (TencentLoginError, TimeoutError) as e:
                continue
            # UserBreak, SystemExit: raise as is

        if self.strategy == "forbid":
            msg = "您可能被限制账密登陆. 扫码登陆仍然可行."
        elif self.strategy != "force":
            msg = "您可能已被限制登陆."
        else:
            msg = "你在睡觉！"

        self.add_hook_ref("hook", self.hook.LoginFailed(msg))
        raise LoginError(msg, self.strategy)


QrStrategy = MixedLoginMan.QrStrategy



import time
from datetime import date, datetime, timedelta
from typing import Optional, Union

from pytz import timezone

TIME_ZONE = timezone("Asia/Shanghai")


def time_ms(ts: Optional[float] = None):
    """int timestamp in ms.

    Args:
        ts (float, optional): timestamp in seconds. Defaults to `time.time()`.

    Returns:
        int: ms timestamp
    """
    return round((ts or time.time()) * 1000)


def dayspac(ts1: float, ts2: Optional[float] = None):
    """return ts2 - ts1

    Args:
        ts1 (float): past timestamp, in seconds, like `time.time()`
        ts2 (float, optional): timestamp, in seconds. Defaults to `time.time()`.

    Returns:
        float: dayspac in float
    """
    d = date.fromtimestamp(ts2 or time.time()) - date.fromtimestamp(ts1)
    return d.seconds / 86400


def approx_ts(timedesc: str) -> int:
    """Given a time description, returns an approximate timestamp.

    :param timedesc: time description (zh-CN)
    :return: timestamp (approximate)
    """
    timedesc = timedesc.strip()
    assert timedesc
    hashm = ":" in timedesc
    didx = timedesc.find("天")
    hmfmt = "%H:%M"

    if "日" in timedesc:
        # specific day
        fmt = "%Y年%m月%d日"
        dt = datetime.strptime(timedesc, fmt + hmfmt if hashm else fmt)
        return int(dt.replace(tzinfo=TIME_ZONE).timestamp())

    today = datetime.today().astimezone(TIME_ZONE)  # based on today
    today = today.replace(second=0, microsecond=0)
    if hashm:
        tm = datetime.strptime(timedesc[-5:], hmfmt).time()
        dt = today.replace(hour=tm.hour, minute=tm.minute)
        if didx == -1:
            return int(dt.timestamp())

        # day in delta
        assert didx > 0
        delta = timedelta(days={"昨": 1, "前": 2}[timedesc[didx - 1]])
        dt -= delta
        return int(dt.timestamp())

    # today
    dt = time.strptime(timedesc.strip(), hmfmt)
    return int(time.mktime(dt))


def sementic_time(timestamp: Union[float, int]) -> str:
    """reverse of :meth:`.approx_ts`

    :param timestamp: timestamp in second
    :return: a sementic time description in Chinese
    """
    today = datetime.now(TIME_ZONE).replace(hour=0, second=0, microsecond=0)
    ytday = today - timedelta(days=1)
    byday = ytday - timedelta(days=1)

    feedtime = datetime.fromtimestamp(timestamp, TIME_ZONE)
    s = ""
    if today <= feedtime:
        pass
    elif ytday <= feedtime < today:
        s += "昨天"
    elif byday <= feedtime < ytday:
        s += "前天"
    else:
        s += feedtime.strftime("%m月%d日 ")
    s += feedtime.strftime("%H:%M")
    return s



"""
Define hooks that can trigger user actions.
"""

import asyncio
from collections import defaultdict
from itertools import chain
from typing import Any, Callable, Coroutine, Dict, Generic, Optional, Set, Tuple, TypeVar


class Event:
    """Base class for event system."""

    pass


Evt = TypeVar("Evt", bound=Event)
T = TypeVar("T")


class NullEvent(Event):
    """For debugging"""

    __slots__ = ()

    def __getattribute__(self, __name: str):
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            raise AssertionError("call `o.register_hook` before accessing o.hook")


class Emittable(Generic[Evt]):
    """An object has some event to trigger."""

    hook: Evt = NullEvent()  # type: ignore
    _tasks: Dict[str, Set[asyncio.Task]]
    _loop: asyncio.AbstractEventLoop

    def __init__(self) -> None:
        self._tasks = defaultdict(set)
        self._loop = asyncio.get_event_loop()

    def register_hook(self, hook: Evt):
        assert not isinstance(hook, NullEvent)
        self.hook = hook

    def add_hook_ref(self, hook_cls, coro):
        # type: (str, Coroutine[Any, Any, T]) -> asyncio.Task[T]
        # NOTE: asyncio.Task becomes generic since py39
        task = self._loop.create_task(coro)  # type: ignore
        self._tasks[hook_cls].add(task)
        task.add_done_callback(lambda t: self._tasks[hook_cls].remove(t))
        return task  # type: ignore

    async def wait(
        self,
        *hook_cls: str,
        timeout: Optional[float] = None,
    ) -> Tuple[Set[asyncio.Task], Set[asyncio.Task]]:
        """Wait for all task in the specific task set(s).

        :param timeout: timeout, defaults to None
        :type timeout: float, optional
        :return: done, pending

        .. note::
            If `timeout` is None, this method will loop until all tasks in the set(s) are done.
            That means if other tasks are added during awaiting, the added task will be waited as well.

        .. seealso:: :meth:`asyncio.wait`
        """
        s = set()
        for i in hook_cls:
            s.update(self._tasks[i])
        if not s:
            return set(), set()
        r = await asyncio.wait(s, timeout=timeout)
        if timeout is None and any(self._tasks[i] for i in hook_cls):
            # await potential new tasks in these sets, only if no timeout.
            r2 = await Emittable.wait(self, *hook_cls)
            return set(chain(r[0], r2[0])), set()
        return r

    def clear(self, *hook_cls: str, cancel: bool = True):
        """Clear the given task sets

        :param hook_cls: task class names
        :param cancel: Cancel the task if a set is not empty, defaults to True. Else will just clear the ref.
        """
        for i in hook_cls:
            s = self._tasks[i]
            if s and not cancel:
                s.clear()
                continue
            while s:
                t = s.pop()
                t.cancel()


class LoginEvent(Event):
    """Defines usual events happens during login."""

    async def LoginFailed(self, msg: Optional[str] = None):
        """Will be emitted on login failed.

        .. note::
            This event will be triggered on every failure of login attempt.
            Means that logic in this event may be executed multiple times.

        :param msg: Err msg, defaults to None.
        """
        pass

    async def LoginSuccess(self):
        """Will be emitted after login success. Low prior, scheduled by loop instead of awaiting."""
        pass


class QREvent(LoginEvent):
    """Defines usual events happens during QR login."""

    cancel: Optional[Callable[[], Coroutine[Any, Any, None]]] = None
    resend: Optional[Callable[[], Coroutine[Any, Any, None]]] = None

    async def QrFetched(self, png: bytes, renew: bool = False):
        """Will be called on new QR code bytes are fetched. Means this will be triggered on:

        1. QR login start
        2. QR expired
        3. QR is refreshed

        :param png: QR bytes (png format)
        :param renew: this QR is a refreshed QR, defaults to False

        .. deprecated:: ?

            Develpers had better not rely on the parameter `renew`.
            Maintaining the state by yourself is not difficult.
        """
        pass

    async def QrFailed(self, msg: Optional[str] = None):
        """QR login failed.

        .. note:: This event should always be called before :meth:`.LoginEvent.LoginFailed`.

        :param msg: Error msg, defaults to None.
        """
        pass

    async def QrSucceess(self):
        """QR login success."""
        pass



"""This module defines types that are used internally by aioqzone and its plugins.
These types are not designed to represent responses from Qzone.
"""

from pydantic import BaseModel


class PersudoCurkey(str):
    def __new__(cls, uin: int, abstime: int):
        return str.__new__(cls, cls.build(uin, abstime))

    def __init__(self, uin: int, abstime: int) -> None:
        super().__init__()
        self.uin = uin
        self.abstime = abstime

    @classmethod
    def build(cls, uin: int, abstime: int):
        return str(uin).rjust(12, "0") + str(abstime).rjust(12, "0")

    @classmethod
    def from_str(cls, curkey: str):
        uin = curkey[:12]
        abstime = curkey[12:]
        uin = int(uin.lstrip("0"))
        abstime = int(abstime.lstrip("0"))
        return cls(uin=uin, abstime=abstime)


class AlbumData(BaseModel):
    topicid: str
    pickey: str
    hostuin: int


# LikeData is not a response of any API. It's just a type def.
class LikeData(BaseModel):
    unikey: str
    curkey: str
    appid: int
    typeid: int
    fid: str
    abstime: int

    @staticmethod
    def persudo_curkey(uin: int, abstime: int):
        return str(PersudoCurkey(uin, abstime))

    @staticmethod
    def persudo_unikey(appid: int, uin: int, **kwds):
        if appid == 311:
            fid = kwds.get("fid", None) or kwds.get("key")
            return f"https://user.qzone.qq.com/{uin}/mood/{fid}"

        raise ValueError(appid)



"""This module defines feed text entities. Entities take place of "contents" since this simplifies
and details the works of rich-text stringify.

.. versionadded:: 0.5.0.dev0
"""

from typing import List, Optional, Union

from pydantic import BaseModel, Field


class HasContent(BaseModel):
    content: str = ""


class ConEntity(BaseModel):
    """Base class for all content entities. All entities has a #type."""

    type: int


class TextEntity(ConEntity):
    """Text, type=2."""

    type: int = 2
    con: str = ""


class AtEntity(ConEntity):
    """At, type=0. like: ``@bot``"""

    type: int = 0
    nick: str = ""
    uin: int


class EmEntity(ConEntity):
    """Emoji, type=-1. like: ``[em]e100[/em]``"""

    type: int = -1
    eid: int


RespEntities = Union[AtEntity, TextEntity]
"""Entities that occurs in Qzone responses."""


class HasConEntity(HasContent):
    entities: Optional[List[RespEntities]] = Field(default=None, alias="conlist")



"""Augment python `dict` since we cannot use py37+ features...

Maybe these are user-defined syntax-sugars?
"""

from itertools import chain
from typing import Dict, Hashable, Iterable, Mapping, MutableMapping, Tuple, TypeVar

K = TypeVar("K", bound=Hashable)
V = TypeVar("V")


def u_(*dicts: Mapping[K, V]) -> Iterable[Tuple[K, V]]:
    return chain(*(i.items() for i in dicts))


def ud(*dicts: Mapping[K, V]) -> Dict[K, V]:
    """returns union of these dict."""
    return dict(u_(*dicts))


def di(d: MutableMapping[K, V], **kw) -> MutableMapping[K, V]:
    """insert kw into this dict."""
    d.update(**kw)
    return d



from typing import Optional


class QzoneError(RuntimeError):
    """HTTP OK, but Qzone returns an error code."""

    msg = "unknown"

    def __init__(self, code: int, *args, rdict=None):
        self.code = int(code)
        self.rdict = rdict
        if len(args) > 0 and isinstance(args[0], str):
            self.msg = args[0]
        RuntimeError.__init__(self, *args)

    def __str__(self) -> str:
        return f"Code {self.code}: {self.msg}"


class LoginError(RuntimeError):
    """Login failed for some reasons."""

    def __init__(self, msg: str, strategy: Optional[str] = None) -> None:
        msg = "登陆失败: " + msg
        super().__init__(msg, strategy)
        self.msg = msg
        self.strategy = strategy


class CorruptError(ValueError):
    """Data corrupted in transfer."""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)



import re
from dataclasses import dataclass
from random import choice, random
from time import time_ns
from typing import Dict, Optional, Type

from aiohttp import ClientSession
from multidict import istr

from ..base import LoginBase
from ..constants import StatusCode
from ..exception import TencentLoginError
from ..type import APPID, PT_QR_APP, CheckResult, Proxy
from ..utils import get_all_cookie
from .encrypt import PasswdEncoder, TeaEncoder

CHECK_URL = "https://ssl.ptlogin2.qq.com/check"
LOGIN_URL = "https://ssl.ptlogin2.qq.com/login"
LOGIN_JS = "https://qq-web.cdn-go.cn/any.ptlogin2.qq.com/v1.3.0/ptlogin/js/c_login_2.js"


@dataclass
class User:
    uin: int
    pwd: str


class UPLogin(LoginBase):
    node = "node"
    _captcha = None
    encode_cls: Type[PasswdEncoder] = TeaEncoder

    def __init__(
        self,
        sess: ClientSession,
        app: APPID,
        proxy: Proxy,
        user: User,
        info: Optional[PT_QR_APP] = None,
    ):
        super().__init__(sess, app, proxy, info=info)
        assert user.uin
        assert user.pwd
        self.user = user
        self.pwder = self.encode_cls(sess, user.pwd)

    async def encodePwd(self, r: CheckResult) -> str:
        return await self.pwder.encode(r)

    async def check(self):
        """[summary]

        Raises:
            HTTPError: [description]

        Returns:
            dict:
                code = 0/2/3 hideVC;
                code = 1 showVC
        """
        data = {
            "regmaster": "",
            "pt_tea": 2,
            "pt_vcode": 1,
            "uin": self.user.uin,
            "appid": self.app.appid,
            # 'js_ver': 21072114,
            "js_type": 1,
            "login_sig": self.login_sig,
            "u1": self.proxy.s_url,
            "r": random(),
            "pt_uistyle": 40,
        }
        async with self.session.get(CHECK_URL, params=data, ssl=self.ssl) as r:
            r.raise_for_status()
            r = re.findall(r"'(.*?)'[,\)]", await r.text())
        r[0] = int(r[0])
        return CheckResult(*r)

    async def sms(self, pt_sms_ticket: str):
        data = {
            "bkn": "",
            "uin": self.user.uin,
            "aid": self.app.appid,
            "pt_sms_ticket": pt_sms_ticket,
        }
        async with self.session.get(
            "https://ui.ptlogin2.qq.com/ssl/send_sms_code", params=data, ssl=self.ssl
        ) as r:
            rl = re.findall(r"'(.*?)'[,\)]", await r.text(encoding="utf8"))
            # ptui_sendSMS_CB('10012', '短信发送成功！')
        assert int(rl[0]) == 10012, rl[1]

    async def login(self, r: CheckResult, pastcode: int = 0) -> Dict[str, str]:
        if r.code == StatusCode.Authenticated:
            # OK
            pass
        elif r.code == StatusCode.NeedCaptcha and pastcode == 0:
            # 0 -> 1: OK; !0 -> 1: Error
            cookie = await self.login(await self.passVC(r), StatusCode.NeedCaptcha)
            return cookie
        elif r.code == StatusCode.NeedVerify and pastcode == StatusCode.NeedVerify:
            # !10009 -> 10009: OK; 10009 -> 10009: Error
            raise TencentLoginError(r.code, str(r))
        else:
            raise TencentLoginError(r.code, str(r))

        data = {
            "u": self.user.uin,
            "p": await self.encodePwd(r),
            "verifycode": r.verifycode,
            "pt_vcode_v1": 1 if pastcode == StatusCode.NeedCaptcha else 0,
            "pt_verifysession_v1": r.verifysession,
            "pt_randsalt": r.isRandSalt,
            "u1": self.proxy.s_url,
            "ptredirect": 0,
            "h": 1,
            "t": 1,
            "g": 1,
            "from_ui": 1,
            "ptlang": 2052,
            "action": f"{3 if pastcode == StatusCode.NeedCaptcha else 2}-{choice([1, 2])}-{int(time_ns() / 1e6)}",
            # 'js_ver': 21072114,
            "js_type": 1,
            "login_sig": self.login_sig,
            "pt_uistyle": 40,
            "aid": self.app.appid,
            "daid": self.app.daid,
            "ptdrvs": r.ptdrvs,
            "sid": r.session,
        }
        self.session.headers.update({istr("referer"): "https://xui.ptlogin2.qq.com/"})
        async with self.session.get(LOGIN_URL, params=data, ssl=self.ssl) as response:
            response.raise_for_status()
            rl = re.findall(r"'(.*?)'[,\)]", await response.text())

        rl[0] = int(rl[0])
        if rl[0] == StatusCode.Authenticated:
            pass
        elif rl[0] == StatusCode.NeedVerify:
            m = response.cookies.get("pt_sms_ticket")
            raise NotImplementedError
            await self.sms(m.value if m else "")
        else:
            raise TencentLoginError(rl[0], rl[4])

        async with self.session.get(rl[2], allow_redirects=False, ssl=self.ssl) as response:
            return get_all_cookie(response)

    def captcha(self, sid: str):
        if not self._captcha:
            from .captcha import Captcha

            self._captcha = Captcha(self.session, self.ssl, self.app.appid, sid)
        return self._captcha

    async def passVC(self, r: CheckResult):
        c = self.captcha(r.session)
        await c.prehandle(self.xlogin_url)
        d = await c.verify()
        r.code = int(d["errorCode"])
        r.verifycode = d["randstr"]
        r.verifysession = d["ticket"]
        return r



"""This module implements (or calls) qzone password encrypt algorithm."""

import base64
import binascii
import hashlib
import re
import struct
from abc import ABC, abstractmethod
from random import randint
from typing import Awaitable

import rsa
from aiohttp import ClientSession

from ..type import CheckResult

LOGIN_JS = "https://qq-web.cdn-go.cn/any.ptlogin2.qq.com/v1.3.0/ptlogin/js/c_login_2.js"


class PasswdEncoder(ABC):
    def __init__(self, sess: ClientSession, passwd: str) -> None:
        self.sess = sess
        self._passwd = passwd

    def encode(self, r: CheckResult) -> Awaitable[str]:
        assert self._passwd, "password should not be empty"
        return self.getEncryption(r.salt, r.verifycode)

    async def login_js(self):
        async with self.sess.get(LOGIN_JS) as response:
            response.raise_for_status()
            return await response.text()

    @abstractmethod
    async def getEncryption(self, salt: str, verifycode: str) -> str:
        pass


class NodeEncoder(PasswdEncoder):
    """
    AGPL-3.0

    Copyright (C) 2021-2022 aioqzone

    .. seealso:: https://github.com/JamzumSum/QQQR/blob/master/LICENCE
    """

    __f = None

    async def getEncryption(self, salt: str, verifycode: str) -> str:
        if self.__f is None:
            js = await self.login_js()
            m = re.search(r"function\(module,exports,__webpack_require__\).*\}", js)
            assert m
            funcs = m.group(0)
            js = "var navigator = new Object; navigator.appName = 'Netscape';"
            js += f"var a=[{funcs}];"
            js += "function n(k) {var t,e=new Object;return a[k](t,e,n),e}\n"
            js += "function getEncryption(p,s,v){var t,e=new Object;return a[9](t,e,n),e['default'].getEncryption(p,s,v,undefined)}"

            from jssupport.execjs import ExecJS

            js = ExecJS("node", js=js)
            self.__f = js.bind("getEncryption")

        return (await self.__f(self._passwd, salt, verifycode)).strip()


class TeaEncoder(PasswdEncoder):
    """The MIT License

    Copyright (c) 2005 hoxide

    Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
    The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    QQ Crypt module.

    .. seealso:: https://github.com/LeoHuang2015/qqloginjs/blob/7d82f2f7d7363547763c40ce5d258d18989b9732/tea.py"""

    op = 0xFFFFFFFF
    delta = 0x9E3779B9
    rsaE = "e9a815ab9d6e86abbf33a4ac64e9196d5be44a09bd0ed6ae052914e1a865ac8331fed863de8ea697e9a7f63329e5e23cda09c72570f46775b7e39ea9670086f847d3c9c51963b131409b1e04265d9747419c635404ca651bbcbc87f99b8008f7f5824653e3658be4ba73e4480156b390bb73bc1f8b33578e7a4e12440e9396f2552c1aff1c92e797ebacdc37c109ab7bce2367a19c56a033ee04534723cc2558cb27368f5b9d32c04d12dbd86bbd68b1d99b7c349a8453ea75d1b2e94491ab30acf6c46a36a75b721b312bedf4e7aad21e54e9bcbcf8144c79b6e3c05eb4a1547750d224c0085d80e6da3907c3d945051c13c7c1dcefd6520ee8379c4f5231ed"
    rsaPublickey = int(rsaE, 16)
    rsaKey = rsa.PublicKey(rsaPublickey, int("10001", 16))

    def _xor(self, a, b):
        a1, a2 = struct.unpack(">LL", a[0:8])
        b1, b2 = struct.unpack(">LL", b[0:8])
        r = struct.pack(">LL", (a1 ^ b1) & self.op, (a2 ^ b2) & self.op)
        return r

    def _tea(self, v: bytes, t: bytes):
        n = 16  # qq use 16
        k = struct.unpack(">LLLL", t[0:16])
        y, z = struct.unpack(">LL", v[0:8])
        s = 0
        for i in range(n):
            s += self.delta
            y += (self.op & (z << 4)) + k[0] ^ z + s ^ (self.op & (z >> 5)) + k[1]
            y &= self.op
            z += (self.op & (y << 4)) + k[2] ^ y + s ^ (self.op & (y >> 5)) + k[3]
            z &= self.op
        r = struct.pack(">LL", y, z)
        return r

    def encrypt(self, v: bytes, k: bytes) -> bytes:
        END_CHAR = b"\0"
        FILL_N_OR = 0xF8

        vl = len(v)
        filln = (8 - (vl + 2)) % 8 + 2
        fills = bytes(randint(0, 0xFF) for _ in range(filln))
        v = bytes([(filln - 2) | FILL_N_OR]) + fills + v + END_CHAR * 7
        tr = b"\0" * 8
        to = b"\0" * 8
        r = b""
        o = b"\0" * 8
        # print 'len(v)=', len(v)
        for i in range(0, len(v), 8):
            o = self._xor(v[i : i + 8], tr)
            tr = self._xor(self._tea(o, k), to)
            to = o
            r += tr
        return r

    def _tx_md5(self, raw_str: bytes):
        return hashlib.md5(raw_str).hexdigest().upper()

    def _hex2str(self, ct: int):
        return hex(ct)[2:]

    def _fromhex(self, s: str):
        return bytes(bytearray.fromhex(s))

    async def getEncryption(self, salt: str, verifycode: str):
        e = salt.encode()
        # md5_pwd = o = self.tx_md5(pwd.encode())
        r = hashlib.md5(self._passwd.encode()).digest()
        p = self._tx_md5(r + e)
        a = rsa.encrypt(r, self.rsaKey)
        rsaData = binascii.b2a_hex(a)

        # rsa length
        s = self._hex2str(int(len(rsaData) / 2))
        s = s.zfill(4)

        # verifycode先转换为大写，然后转换为bytes
        l = binascii.b2a_hex(verifycode.upper().encode())

        # verifycode length
        c = self._hex2str(int(len(l) / 2))
        c = c.zfill(4)

        # TEA: KEY:p, s+a+ TEA.strToBytes(e) + c +l
        # aioqzone EDIT: I tried `binascii.b2a_hex(e)` to replace `salt` to fix error
        new_pwd = s.encode() + rsaData + binascii.b2a_hex(e) + c.encode() + l
        enc = self.encrypt(self._fromhex(new_pwd.decode()), self._fromhex(p))
        return base64.b64encode(enc, b"*-").decode().replace("=", "_")



import asyncio
from abc import ABC, abstractmethod
from time import time
from typing import Dict

from qqqr.encrypt import gtk

from ..interface.hook import Emittable, LoginEvent


class Loginable(ABC, Emittable[LoginEvent]):
    last_login: float = 0

    def __init__(self, uin: int) -> None:
        super().__init__()
        self.uin = uin
        self._cookie = {}
        self.lock = asyncio.Lock()

    @property
    def cookie(self) -> Dict[str, str]:  # type: ignore
        """Get cookie in any way. Allow cached result.

        Returns:
            int: cookie. Cached cookie is preferable.
        """
        return self._cookie

    @abstractmethod
    async def _new_cookie(self) -> Dict[str, str]:
        return

    async def new_cookie(self):
        """Get a new cookie. Means, cached cookie is not allowed.

        Returns:
            int: cookie. Shouldn't be a cached one.
        """
        if self.lock.locked():
            # if there is coro. updating cookie, then wait for its result.
            async with self.lock:
                return self.cookie
        else:
            # let the first coro. get result from Qzone.
            async with self.lock:
                self._cookie = await self._new_cookie()
                self.last_login = time()
                return self._cookie

    @property
    def gtk(self) -> int:
        """cal gtk from pskey.

        Returns:
            int: gtk. NOTE: 0 denotes no-login.
        """
        pskey = self.cookie.get("p_skey")
        if pskey is None:
            return 0
        return gtk(pskey)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aioqzone',
 'aioqzone.api',
 'aioqzone.interface',
 'aioqzone.type',
 'aioqzone.utils',
 'jssupport',
 'qqqr',
 'qqqr.qr',
 'qqqr.up',
 'qqqr.up.captcha']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'cssselect>=1.1.0,<2.0.0',
 'lxml>=4.8.0,<5.0.0',
 'opencv-python-headless>=4.5.5,<5.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pytz>=2022.1,<2023.0',
 'rsa>=4.8,<5.0']

extras_require = \
{'dev': ['tox>=3.24.5,<4.0.0',
         'pre-commit>=2.17.0,<3.0.0',
         'black>=22.1.0,<23.0.0',
         'isort>=5.10.1,<6.0.0'],
 'doc': ['Sphinx>=4.5.0,<5.0.0',
         'autodoc-pydantic>=1.6.1,<2.0.0',
         'sphinx-autodoc-typehints<1.16.0',
         'sphinx-rtd-theme>=1.0.0,<2.0.0']}

setup_kwargs = {
    'name': 'aioqzone',
    'version': '0.5.1a1',
    'description': 'Python wrapper for Qzone web login and Qzone http api.',
    'long_description': '# aioqzone\n\naioqzone is a python package handling Qzone web login and wrapping some common Qzone Http apis.\n\n[![black](https://img.shields.io/badge/python-3.7%20%7C%203.10-blue)][home]\n[![QQQR](https://github.com/aioqzone/aioqzone/actions/workflows/qqqr.yml/badge.svg?branch=beta&event=schedule)](https://github.com/aioqzone/aioqzone/actions/workflows/qqqr.yml)\n[![black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n[简体中文](README_zh-cn.md)\n\n## Features\n\n### Qzone Feature\n\n- [x] QR login\n- [x] password login (limited)\n- [ ] passing captcha (implemented but seems not working...)\n- [x] get complete html feeds\n- [x] get feed details\n- [x] get Qzone album\n- [x] like/unlike app\n- [x] publish/update/delete text feeds\n- [ ] comment\n\n### Why using this package?\n\n- [x] full ide typing support (typing)\n- [x] api response validation (pydantic)\n- [x] async design\n- [x] complete infrastructure to ease your own develop\n- [x] [doc support](https://aioqzone.github.io/aioqzone)\n\n__Working On:__\n\n- [ ] test coverage\n\n## Dependency out of `setup.cfg`\n\n- NodeJS (jssupport)\n\n## Description\n\n|package    |brief description  |\n|-----------|-------------------|\n|aioqzone   |qzone api wrapper  |\n|jssupport  |exec js            |\n|qqqr       |qzone web login    |\n\n## Examples\n\nYou can look for these repos for examples in practice.\n\n### aioqzone plugins\n\n- [aioqzone-feed][aioqzone-feed]: aioqzone plugin providing higher level api for processing feed\n\n\n## License\n\n```\nCopyright (C) 2022 aioqzone.\n\nThis program is free software: you can redistribute it and/or modify\nit under the terms of the GNU Affero General Public License as published\nby the Free Software Foundation, either version 3 of the License, or\n(at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU Affero General Public License for more details.\n\nYou should have received a copy of the GNU Affero General Public License\nalong with this program.  If not, see <https://www.gnu.org/licenses/>.\n```\n\n- [AGPL-3.0](LICENSE)\n\n\n[home]: https://github.com/aioqzone/aioqzone "Python wrapper for Qzone web login and Qzone http api"\n[aioqzone-feed]: https://github.com/aioqzone/aioqzone-feed "aioqzone plugin providing higher level api for processing feed"\n',
    'author': 'aioqzone',
    'author_email': 'zzzzss990315@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aioqzone/aioqzone',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



"""This module defines types that represent responses from Qzone.
These `*Rep` classes are all :class:`pydantic.BaseModel`, and can be used to validate responses."""

from typing import List, Optional, Union

from pydantic import BaseModel, Field
from pydantic.networks import HttpUrl

from .entity import HasConEntity, HasContent


# Below are the response reps of Qzone Apis.
class FeedRep(BaseModel):
    ver: int
    appid: int
    typeid: int
    fid: str = Field(alias="key")
    abstime: int
    uin: int
    nickname: str
    html: str
    # likecnt: Optional[int] = None
    # relycnt: Optional[int] = None
    # commentcnt: Optional[int] = None


class FeedMoreAux(BaseModel):
    hasMoreFeeds: bool
    dayspac: int
    pagenum: int
    externparam: str
    begintime: int
    endtime: str

    lastaccesstime: str
    lastAccessRelateTime: str

    aisortBeginTime: str
    aisortEndTime: str
    aisortOffset: str
    aisortNextTime: str

    daylist: str
    uinlist: str

    # attach: str
    # searchtype: str
    # error: str
    # hotkey: str
    # icGroupData: list
    # host_level: str
    # friend_level: str
    # hidedNameList: list
    # owner_bitmap: str


class CommentRep(HasContent):
    abstime: int = Field(alias="create_time")
    name: str
    uin: int
    replyNum: int
    tid: int
    pic: Optional[List[dict]] = None


class PicRep(BaseModel):
    height: int = 0
    width: int = 0
    thumb: Union[HttpUrl, str] = Field(alias="url1")
    picId: Union[HttpUrl, str] = Field(alias="url2")
    raw: Union[HttpUrl, str] = Field(alias="url3")
    is_video: int = False

    def from_url(self, url: HttpUrl):
        self.thumb = self.picId = self.raw = url

    @classmethod
    def from_floatview(cls, fv: "FloatViewPhoto"):
        if fv.is_video and fv.video_info:
            return VideoRep.from_floatview(fv)
        else:
            return cls(
                height=fv.height,
                width=fv.width,
                is_video=False,
                url1=fv.thumb,
                url2=fv.picId,
                url3=fv.raw,
            )

    def valid_url(self):
        """Test if `self.thumb` and `self.raw` is :class:`HttpUrl`."""
        return isinstance(self.thumb, HttpUrl) and isinstance(self.raw, HttpUrl)


class VideoInfo(BaseModel):
    cover_height: int = 0  # same as VideoRep.height
    cover_width: int = 0
    thumb: HttpUrl = Field(alias="url1")
    raw: HttpUrl = Field(alias="url3")
    vid: str = Field(alias="video_id")


class VideoInfo2(BaseModel):
    cover_height: int = 0  # same as VideoRep.height
    cover_width: int = 0
    vid: str
    raw: HttpUrl = Field(alias="video_url")


class VideoRep(PicRep):
    video_info: VideoInfo

    @classmethod
    def from_floatview(cls, fv: "FloatViewPhoto"):
        assert fv.is_video
        assert fv.video_info
        return cls(
            height=fv.height,
            width=fv.width,
            is_video=True,
            url1=fv.thumb,
            url2=fv.picId,
            url3=fv.raw,
            video_info=VideoInfo(
                cover_height=fv.height,
                cover_width=fv.width,
                url1=fv.thumb,
                url3=fv.video_info.raw,
                video_id=fv.video_info.vid,
            ),
        )


class FeedDetailRep(HasConEntity):
    uin: int
    name: str
    fid: str = Field(alias="tid")
    abstime: int = Field(alias="created_time")

    # forward from
    rt_con: Optional[HasConEntity] = None
    rt_uin: int = 0
    rt_uinname: str = ""
    rt_fid: str = Field(default="", alias="rt_tid")
    rt_createTime: str = ""
    rt_abstime: int = Field(default=0, alias="rt_created_time")
    pic: Optional[List[Union[VideoRep, PicRep]]] = None

    cmtnum: int
    commentlist: Optional[List[CommentRep]] = None
    fwdnum: int


class FeedsCount(BaseModel):
    aboutHostFeeds_new_cnt: int
    replyHostFeeds_new_cnt: int
    myFeeds_new_cnt: int
    friendFeeds_new_cnt: int
    friendFeeds_newblog_cnt: int
    friendFeeds_newphoto_cnt: int
    specialCareFeeds_new_cnt: int
    followFeeds_new_cnt: int
    newfeeds_uinlist: list


class FloatViewPhoto(BaseModel):
    albumId: str
    createTime: str
    desc: dict
    ownerName: str
    ownerUin: int
    photoOwner: int
    fid: str = Field(alias="tid")
    topicId: str

    height: int
    width: int

    thumb: HttpUrl = Field(alias="pre")
    raw: HttpUrl = Field(alias="url")
    picId: Union[HttpUrl, str]
    picKey: str
    video_info: Optional[VideoInfo2] = None

    cmtTotal: int
    fwdnum: int

    likeTotal: int
    likeKey: str
    likeList: Optional[List[dict]] = None

    lloc: str
    rt_fid: str = Field(alias="original_tid")
    photocubage: Optional[int] = None
    phototype: int = 1

    isMultiPic: Optional[bool] = False
    is_weixin_mode: Optional[bool] = False
    is_video: Optional[bool] = False


class MsgListElm(HasContent):
    cmtnum: int
    fwdnum: int

    createTime: str
    abstime: int = Field(alias="created_time")
    commentlist: list

    fid: str = Field(alias="tid")
    uin: int
    name: str



import asyncio
import re
from random import random
from typing import Awaitable, Callable, Dict, Optional

from ..base import LoginBase
from ..constants import StatusCode
from ..encrypt import hash33
from ..exception import UserBreak
from ..utils import get_all_cookie, raise_for_status

SHOW_QR = "https://ssl.ptlogin2.qq.com/ptqrshow"
XLOGIN_URL = "https://xui.ptlogin2.qq.com/cgi-bin/xlogin"
POLL_QR = "https://ssl.ptlogin2.qq.com/ptqrlogin"
LOGIN_URL = "https://ptlogin2.qzone.qq.com/check_sig"


class QRLogin(LoginBase):
    async def show(self):
        data = {
            "appid": self.app.appid,
            "e": 2,
            "l": "M",
            "s": 3,
            "d": 72,
            "v": 4,
            "t": random(),
            "daid": self.app.daid,
            "pt_3rd_aid": 0,
        }
        async with self.session.get(SHOW_QR, params=data, ssl=self.ssl) as r:
            r.raise_for_status()
            self.qrsig = r.cookies["qrsig"].value
            return await r.content.read()

    async def pollStat(self):
        """poll status of the qr

        Raises:
            HTTPError: if response code != 200

        Returns:
            list: (code, ?, url, ?, msg, my_name)
        """
        data = {
            "u1": self.proxy.s_url,
            "ptqrtoken": hash33(self.qrsig),
            "ptredirect": 0,
            "h": 1,
            "t": 1,
            "g": 1,
            "from_ui": 1,
            "ptlang": 2052,
            # 'action': 3-2-1626611307380,
            # 'js_ver': 21071516,
            "js_type": 1,
            "login_sig": "",
            "pt_uistyle": 40,
            "aid": self.app.appid,
            "daid": self.app.daid,
            # 'ptdrvs': 'JIkvP2N0eJUzU3Owd7jOvAkvMctuVfODUMSPltXYZwCLh8aJ2y2hdSyFLGxMaH1U',
            # 'sid': 6703626068650368611,
            "has_onekey": 1,
        }
        async with self.session.get(POLL_QR, params=data, ssl=self.ssl) as r:
            r.raise_for_status()
            r = re.findall(r"'(.*?)'[,\)]", await r.text())
            r[0] = int(r[0])
            if r[0] == 0:
                self.login_url = r[2]
            return r

    async def login(self):
        async with self.session.get(self.login_url, allow_redirects=False, ssl=self.ssl) as r:
            raise_for_status(r, 302)
            return get_all_cookie(r)

    def loop(self, send_callback, expire_callback=None, refresh_time=6, polling_freq=3):
        # type: (Callable[[bytes], Awaitable[None]], Optional[Callable[[bytes], Awaitable[None]]], int, float) -> asyncio.Task[Dict[str, str]]
        expire_callback = expire_callback or send_callback

        async def innerloop():
            assert expire_callback
            try:
                for i in range(refresh_time):
                    send = expire_callback if i else send_callback
                    await send(await self.show())
                    # future.set_exception(UserBreak)
                    while True:
                        await asyncio.sleep(polling_freq)
                        stat = await self.pollStat()
                        if stat[0] == StatusCode.Expired:
                            break
                        if stat[0] == StatusCode.Authenticated:
                            return await self.login()
            except (KeyboardInterrupt, asyncio.CancelledError) as e:
                raise UserBreak from e
            raise TimeoutError

        expire_callback = expire_callback or send_callback
        return asyncio.create_task(innerloop())



"""
Make some easy-to-use api from basic wrappers.
"""
from typing import List, Optional, Tuple

from pydantic import ValidationError

from ..type.internal import AlbumData
from ..type.resp import *
from .raw import QzoneApi


class DummyQapi(QzoneApi):
    """A wrapper of :class:`.QzoneApi`. Validate the responses, returns pydantic `BaseModel`."""

    async def feeds3_html_more(
        self, pagenum: int, trans: Optional[QzoneApi.FeedsMoreTransaction] = None, count: int = 10
    ) -> Tuple[List[FeedRep], FeedMoreAux]:
        r = await super().feeds3_html_more(pagenum, trans=trans, count=count)
        data = r["data"]
        main = r["main"]
        assert isinstance(data, list)
        return [FeedRep.parse_obj(i) for i in data if i], FeedMoreAux.parse_obj(main)

    async def emotion_getcomments(self, uin: int, tid: str, feedstype: int) -> str:
        r = await super().emotion_getcomments(uin, tid, feedstype)
        return str.strip(r["newFeedXML"])  # type: ignore

    async def emotion_msgdetail(self, owner: int, fid: str) -> FeedDetailRep:
        r = await super().emotion_msgdetail(owner, fid)
        return FeedDetailRep.parse_obj(r)

    async def get_feeds_count(self) -> FeedsCount:
        r = await super().get_feeds_count()
        return FeedsCount.parse_obj(r)

    async def floatview_photo_list(self, album: AlbumData, num: int) -> List[FloatViewPhoto]:
        r = await super().floatview_photo_list(album, num)
        try:
            return [FloatViewPhoto.parse_obj(i) for i in r["photos"]]  # type: ignore
        except ValidationError as e:
            # for DEBUG only
            raise e

    async def emotion_msglist(self, uin: int, num: int = 20, pos: int = 0) -> List[MsgListElm]:
        r = await super().emotion_msglist(uin, num, pos)
        return [MsgListElm.parse_obj(i) for i in r]



import os

import cv2 as cv
import numpy as np

debug = False

TC_OPERATION_WIDTH = 279


def frombytes(b, dtype="uint8", flags=cv.IMREAD_COLOR) -> np.ndarray:
    return cv.imdecode(np.asarray(bytearray(b), dtype=dtype), flags=flags)


class Piece:
    def __init__(self, img: np.ndarray) -> None:
        self.mask = img[:, :, 3:]
        self.img = img[:, :, :3] * (self.mask == 255)

    def strip(self):
        if not hasattr(self, "bbox"):
            cont, (hier,) = cv.findContours(
                cv.cvtColor(self.img, cv.COLOR_BGR2GRAY), cv.RETR_TREE, cv.CHAIN_APPROX_SIMPLE
            )
            center = (int(self.img.shape[0] // 2), int(self.img.shape[1] // 2))
            cond = [
                cv.pointPolygonTest(i, center, False) == 1
                and cv.pointPolygonTest(i, (0, 0), False) == -1
                for i in cont
            ]
            cont = [i for i, c in zip(cont, cond) if c]
            if len(cont) == 2:
                hier = [i for i, c in zip(hier, cond) if c]
                cont = [i for i, c in zip(cont, hier) if c[2] == -1]
            assert len(cont) == 1
            self.setContour(cont[0])

        r = self.img[
            self.bbox[1] : self.bbox[1] + self.bbox[3], self.bbox[0] : self.bbox[0] + self.bbox[2]
        ]
        return r

    def setContour(self, cont: list):
        self.cont = cont
        self.bbox = cv.boundingRect(cont)
        self.padding = (
            *self.bbox[:2],
            self.img.shape[1] - self.bbox[2] - self.bbox[0],
            self.img.shape[0] - self.bbox[3] - self.bbox[1],
        )

    def strip_mask(self):
        r = self.mask[
            self.bbox[1] : self.bbox[1] + self.bbox[3], self.bbox[0] : self.bbox[0] + self.bbox[2]
        ]
        return r * self.strip().astype("bool")

    def imitated(self):
        spiece = (self.strip() * 0.3).astype("uint8")
        cont = np.array(
            [[[x - self.bbox[0], y - self.bbox[1]]] for (x, y), in self.cont], dtype="int"
        )
        return cv.drawContours(spiece, [cont], 0, (255, 255, 255), 1)


class Jigsaw:
    def __init__(self, origin: bytes, puzzle: bytes, piece: bytes, top: int) -> None:
        self.top = top
        self.ans = frombytes(origin)
        self.puzzle = frombytes(puzzle)
        self.piece = Piece(frombytes(piece, flags=cv.IMREAD_UNCHANGED))

    @staticmethod
    def save(ans, puzzle, piece, top):
        import yaml

        os.makedirs("data", exist_ok=True)
        ex = len([i for i in os.listdir("data") if i.endswith(".yml")])
        with open(f"data/{ex}.yml", "w") as f:
            yaml.safe_dump({"origin": ans, "puzzle": puzzle, "piece": piece, "top": top}, f)

    @classmethod
    def load(cls, filename):
        import yaml

        with open(filename) as f:
            return cls(**yaml.safe_load(f))

    @property
    def width(self) -> int:
        return self.ans.shape[1]

    @property
    def rate(self):
        # TODO: operation width / 680
        return TC_OPERATION_WIDTH / 680

    @property
    def left(self):
        if not hasattr(self, "_left"):
            self._left = self.solve() - self.piece.padding[0]
        return self._left

    def solve(self) -> int:
        spiece = self.piece.imitated()
        top = self.top + self.piece.padding[1]

        r = cv.matchTemplate(
            self.puzzle[top : top + self.piece.bbox[3]],
            spiece,
            cv.TM_CCOEFF_NORMED,
            mask=self.piece.strip_mask(),
        )
        left = int(np.argmax(r))

        if debug:
            cv.imshow("match", np.tile(r, (200, 1)))
            cv.imshow("spiece", spiece)

            cont = [
                [[x + left - self.piece.padding[0], y + self.top]] for (x, y), in self.piece.cont
            ]
            cont = np.array(cont, dtype="int")
            d = cv.drawContours(self.puzzle, [cont], 0, (0, 0, 255), 2)
            cv.imshow("contour", d)
            cv.waitKey()

        return left



"""
some patterns for matching html and so on.
"""
import re

# use this to match qzone api response
response_callback = re.compile(r"callback\(\s*(\{.*\})\s*\)", re.S | re.I)
# use this to get unikey & curkey of a html
uni_cur_key = re.compile(r'data-unikey="([^"]*)"[^d]*data-curkey="([^"]*)"')
